/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// listes simplement chaînées

#include "prelude.h"

#include "list.h"

list *list_nil = NULL;

list *list_cons(int x, list *n) {
  list *c = malloc(sizeof(struct Cell));
  c->value = x;
  c->next = n;
  return c;
}

list *list_append(list *l1, list *l2) {
  if (l1 == NULL) return l2;
  list *p = l1, *c = l1->next;
  while (c != NULL) {
    p = c;
    c = c->next;
  }
  p->next = l2;
  return l1;
}

list *list_copy(list *l) {
  if (l == NULL) return l;
  list *h = list_cons(l->value, NULL), *p = h;
  for (l=l->next; l != NULL; l=l->next) {
    list *c = list_cons(l->value, NULL);
    p->next = c;
    p = c;
  }
  return h;
}

list *list_concat(list *l1, list *l2) {
  return list_append(l1, list_copy(l2));
}

int list_nth(list *l, int n) {
  assert(n >= 0);
  while (n > 0 && l != NULL) {
    l = l->next;
    n--;
  }
  assert(l != NULL);
  return l->value;
}

int list_length(list *l) {
  int len = 0;
  for (; l != NULL; l = l->next) {
    len++;
  }
  return len;
}

void list_delete(list *l) {
  while (l != NULL) {
    list *p = l;
    l = l->next;
    free(p);
  }
}

void list_print(list *l) {
  printf("[");
  while (l != NULL) {
    printf("%d", l->value);
    l = l->next;
    if (l != NULL) printf(", ");
  }
  printf("]\n");
}

// renvoie la liste lo,lo+1,...,hi-1
list *list_interval(int lo, int hi) {
  list *l = NULL;
  while (lo < hi--) {
    l = list_cons(hi, l);
  }
  return l;
}

list *list_reversal(list *l) {
  list *last = NULL;
  while (l != NULL) {
    list *tmp = l;
    l = l->next;
    tmp->next = last;
    last = tmp;
  };
  return last;
}

int list_random(list *l) {
  assert(l != NULL);
  int r = 0, n = 1;
  while (l != NULL) {
    if (rand() % n == 0) r = l->value;
    l = l->next;
    n++;
  }
  return r;
}

// le lièvre et la tortue

bool list_cyclic(list *l) {
  if (l == NULL) return false;
  list *tortoise = l, *hare = l->next;
  while (tortoise != hare) {
    if (hare == NULL) return false;
    hare = hare->next;
    if (hare == NULL) return false;
    hare = hare->next;
    tortoise = tortoise->next;
  }
  return true;
}
